// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "SessionBase.h"
#include "log_ama.h"

namespace xma {

#define XMA_SESSION_BASE "SessionBase"

///////////////////////////////////////////////////////////////////////////////
SessionBase::SessionBase(IntSessionFactoryPtr sf, const char* logging_name, XmaParameter* params, int param_cnt) : m_logging_name(logging_name), m_sf(sf) {
  m_log = m_sf->GetLog();

  XmaParameter* param;
  if ((param = GetParameter(params, param_cnt, XMA_FILTER_PARAM_LATENCY_LOGGING)) != nullptr) {
    if (*(int32_t*) param->value < 0 || *(int32_t*) param->value > 1) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::SessionBase() Invalid value for latency_logging");
    } else {
      m_latency_logging = *(int32_t*) param->value;
    }
  } else {
    m_latency_logging = XMA_FILTER_LATENCY_LOGGING_DEFAULT;
  }

  m_log->LogMsg(XMA_INFO_LOG, XMA_SESSION_BASE, "SessionBase::SessionBase() %s_handle : %p", m_logging_name, this);
}

///////////////////////////////////////////////////////////////////////////////
int32_t SessionBase::SetLog(LogPtr log) {
  const char* level;
  switch (log->GetLogLevel()) {
  case XMA_EMERGENCY_LOG:
    level = "emergency";
    break;
  case XMA_ALERT_LOG:
    level = "alert";
    break;
  case XMA_CRITICAL_LOG:
    level = "critical";
    break;
  case XMA_ERROR_LOG:
    level = "error";
    break;
  case XMA_WARNING_LOG:
    level = "warning";
    break;
  case XMA_NOTICE_LOG:
    level = "notice";
    break;
  case XMA_INFO_LOG:
    level = "info";
    break;
  case XMA_DEBUG_LOG:
    level = "debug";
    break;
  default:
    level = "unknown";
    break;
  }
  switch (log->GetLogType()) {
  case XMA_LOG_TYPE_NONE:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=none)", level);
    break;
  case XMA_LOG_TYPE_CONSOLE:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=console)", level);
    break;
  case XMA_LOG_TYPE_SYSLOG:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=syslog)", level);
    break;
  case XMA_LOG_TYPE_FILE: {
    std::shared_ptr<XmaLogFile> log_file = std::dynamic_pointer_cast<XmaLogFile>(log);
    if (log_file) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=file, file='%s')", level, log_file->GetFileName());
    } else {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::SetLog() Invalid log!");
    }
    break;
  }
  case XMA_LOG_TYPE_CALLBACK:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=callback)", level);
    break;
  case XMA_LOG_TYPE_AMA:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=AMA)", level);
    break;
  default:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog(level=%s, type=unknown)", level);
    break;
  }
  m_log = log;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::SetLog() - Returning");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
XmaParameter* SessionBase::GetParameter(LogPtr log, XmaParameter* params, int num_params, const char* name) {
  log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::GetParameter(%s)", name);
  int i;
  for (i = 0; i < num_params; ++i) {
    if (!strcmp(name, params[i].name)) {
      log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::GetParameter() - Returning parameter");
      return &params[i];
    }
  }
  log->LogMsg(XMA_DEBUG_LOG, XMA_SESSION_BASE, "SessionBase::GetParameter() - Returning NULL");
  return nullptr;
}

///////////////////////////////////////////////////////////////////////////////
int32_t SessionBase::SetParam(XmaParameter* params, uint32_t param_cnt, const char* param_name, const int32_t min_value, const int32_t max_value,
    const int32_t default_value, const XmaAPIVersion earliest_version, const XmaAPIVersion latest_version) {
  XmaParameter* param;
  if ((param = GetParameter(params, param_cnt, param_name)) != nullptr) {
    const XmaAPIVersion version = m_sf->GetAPIVersion();
    if ((version < earliest_version) || (latest_version < version)) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::SetParam() Unsupported feature (%s) in this version of the API!", param_name);
    }
    int32_t param_value = *reinterpret_cast<int32_t*>(param->value);
    if (param_value < min_value || param_value > max_value) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::SetParam() Invalid value for %s! (range = %d - %d)", param_name, min_value, max_value);
    } else {
      return param_value;
    }
  } else {
    return default_value;
  }
}

///////////////////////////////////////////////////////////////////////////////
void SessionBase::ForEach(LogPtr log, queue_t& queue, std::function<void(IntFramePtr)> frame_func, std::function<void(IntDataBufferPtr)> buf_func) {
  switch (queue.index()) {
  case 0: { // QueueFrame
    FramePtr frame = std::get<QueueFrame>(queue).frame;
    if (frame) {
      IntFramePtr int_frame = std::dynamic_pointer_cast<InternalFrame>(frame);
      if (int_frame) {
        frame_func(int_frame);
      } else {
        THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::ForEach() Invalid frame!");
      }
    }
    break;
  }
  case 1: { // QueueFrameList
    std::vector<FramePtr>& frame_list = std::get<QueueFrameList>(queue).frame_list;
    const size_t           size       = frame_list.size();
    size_t                 index;
    for (index = 0; index < size; ++index) {
      FramePtr frame = frame_list[index];
      if (frame) {
        IntFramePtr int_frame = std::dynamic_pointer_cast<InternalFrame>(frame);
        if (int_frame) {
          frame_func(int_frame);
        } else {
          THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::ForEach() Invalid frame!");
        }
      }
    }
    break;
  }
  case 2: //QueueDataBuffer
    DataBufferPtr buf = std::get<QueueDataBuffer>(queue).data;
    if (buf) {
      IntDataBufferPtr int_buf = std::dynamic_pointer_cast<InternalDataBuffer>(buf);
      if (int_buf) {
        buf_func(int_buf);
      } else {
        THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SESSION_BASE, "SessionBase::ForEach() Invalid data buffer!");
      }
    }
    break;
  }
}

///////////////////////////////////////////////////////////////////////////////
void SessionBase::LogLatencySend(uint64_t frame) {
  if (m_latency_logging && (frame > m_frame_sent)) {
    struct timespec latency;
    clock_gettime(CLOCK_REALTIME, &latency);
    int64_t time_taken = (latency.tv_sec * 1e3) + (latency.tv_nsec / 1e6);
    m_log->LogMsg(XMA_INFO_LOG, XMA_SESSION_BASE, "%p : xma_%s_frame_sent : %lld : %lld\n", this, m_logging_name, frame, time_taken);
  }
  m_frame_sent = frame;
}

///////////////////////////////////////////////////////////////////////////////
void SessionBase::LogLatencyRecv(uint64_t frame) {
  if (m_latency_logging) {
    struct timespec latency;
    clock_gettime(CLOCK_REALTIME, &latency);
    int64_t time_taken = (latency.tv_sec * 1e3) + (latency.tv_nsec / 1e6);
    m_log->LogMsg(XMA_INFO_LOG, XMA_SESSION_BASE, "%p : xma_%s_frame_recv : %lld : %lld\n", this, m_logging_name, frame, time_taken);
  }
}

} // namespace xma
