// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <cstring>
#include "xmaapidecoder.h"
#include "xmaapi.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"
#include "SessionBase.h"

#define XMA_DECODER_MOD "xmadecoder"

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_param_callback(xma::FramePropertiesPtr new_props, void* opaque) {
  XmaDecoderSession* session = reinterpret_cast<XmaDecoderSession*>(opaque);

  xma::IntFramePropertiesPtr int_props = std::static_pointer_cast<xma::InternalFrameProperties>(new_props);
  XmaFrameProperties         props     = int_props->GetFrameProperties();
  return session->m_prop_callback(&props, session->m_prop_param);
}

///////////////////////////////////////////////////////////////////////////////
XmaDecoderSession* xma_dec_session_create(XmaDecoderProperties* dec_props) {
  if (!dec_props) {
    return nullptr;
  }

  XmaSessionFactory* xma_sf = static_cast<XmaSessionFactory*>(dec_props->handle);
  if (!xma_sf) {
    return nullptr;
  }

  TRY()
  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_create()");
  XmaDecoderSession* session = new XmaDecoderSession();
  if (!session) {
    xma_sf->m_sf->GetLog()->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_create() Out Of Memory!");
    return nullptr;
  }

  session->m_sf            = xma_sf->m_sf;
  session->m_log           = xma_sf->m_sf->GetLog();
  session->m_prop_callback = nullptr;
  session->m_prop_param    = nullptr;

  XmaParameter* param_list = nullptr;
  XmaParameter* param      = xma::SessionBase::GetParameter(session->m_log, dec_props->params, dec_props->param_cnt, XMA_DEC_PARAM_PROP_CHANGE_CALLBACK);
  if (param) {
    session->m_prop_callback = reinterpret_cast<DecPropChangeCallbackFunction>(param->value);
    param_list               = dec_props->params;
    dec_props->params        = new XmaParameter[dec_props->param_cnt + 1];
    memcpy(dec_props->params, param_list, sizeof(XmaParameter) * dec_props->param_cnt);
    param        = xma::SessionBase::GetParameter(session->m_log, dec_props->params, dec_props->param_cnt, XMA_DEC_PARAM_PROP_CHANGE_CALLBACK);
    param->value = reinterpret_cast<void*>(xma_dec_param_callback);
    param        = xma::SessionBase::GetParameter(session->m_log, dec_props->params, dec_props->param_cnt, XMA_DEC_PARAM_PROP_CHANGE_PARAM);
    if (param) {
      session->m_prop_param = param->value;
      param->value          = session;
    } else {
      dec_props->params[dec_props->param_cnt].name    = const_cast<char*>(XMA_DEC_PARAM_PROP_CHANGE_PARAM);
      dec_props->params[dec_props->param_cnt].type    = XMA_FUNC_PTR;
      dec_props->params[dec_props->param_cnt].length  = sizeof(void*);
      dec_props->params[dec_props->param_cnt++].value = session;
    }
  }

  session->m_session = xma_sf->m_sf->MakeDecoderSession(dec_props);

  if (param_list) {
    delete[] dec_props->params;
    dec_props->params = param_list;
    if (session->m_prop_param) {
      --dec_props->param_cnt;
      param = xma::SessionBase::GetParameter(session->m_log, dec_props->params, dec_props->param_cnt, XMA_DEC_PARAM_PROP_CHANGE_PARAM);
      if (param) {
        param->value = session->m_prop_param;
      }
    }
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_create() Out Of Memory!");
    delete session;
    return nullptr;
  }

  xma_sf->m_sf->GetLog()->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_create() - Returning");
  return static_cast<XmaDecoderSession*>(session);
  CATCH_NULL(xma_sf->m_sf->GetLog(), XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_destroy(XmaDecoderSession* session) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  xma::LogPtr log = session->m_log;
  log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_destroy()");

  if (!session->m_session) {
    log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_destroy() Invalid parameter!");
    delete session;
    return XMA_ERROR_INVALID;
  }

  TRY()
  delete session;

  log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_destroy() - Returning success");
  return XMA_SUCCESS;
  CATCH_ERROR(log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_set_log(XmaDecoderSession* session, XmaLogHandle handle) {
  if (!session || !handle) {
    return XMA_ERROR_INVALID;
  }

  TRY()
  XmaLogStruct* log_struct = static_cast<XmaLogStruct*>(handle);
  xma::LogPtr   log        = log_struct->m_log;
  const char*   level;
  switch (log->GetLogLevel()) {
  case XMA_EMERGENCY_LOG:
    level = "emergency";
    break;
  case XMA_ALERT_LOG:
    level = "alert";
    break;
  case XMA_CRITICAL_LOG:
    level = "critical";
    break;
  case XMA_ERROR_LOG:
    level = "error";
    break;
  case XMA_WARNING_LOG:
    level = "warning";
    break;
  case XMA_NOTICE_LOG:
    level = "notice";
    break;
  case XMA_INFO_LOG:
    level = "info";
    break;
  case XMA_DEBUG_LOG:
    level = "debug";
    break;
  default:
    level = "unknown";
    break;
  }
  switch (log->GetLogType()) {
  case XMA_LOG_TYPE_NONE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=none)", level);
    break;
  case XMA_LOG_TYPE_CONSOLE:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=console)", level);
    break;
  case XMA_LOG_TYPE_SYSLOG:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=syslog)", level);
    break;
  case XMA_LOG_TYPE_FILE: {
    std::shared_ptr<xma::XmaLogFile> log_file = std::dynamic_pointer_cast<xma::XmaLogFile>(log);
    if (log_file) {
      session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=file, file='%s')", level, log_file->GetFileName());
    } else {
      const char* error_msg = "xma_dec_session_set_log() Invalid log!";
      session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, error_msg);
      throw xma::InvalidArgumentException(error_msg);
    }
    break;
  }
  case XMA_LOG_TYPE_CALLBACK:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=callback)", level);
    break;
  case XMA_LOG_TYPE_AMA:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=AMA)", level);
    break;
  default:
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log(level=%s, type=unknown)", level);
    break;
  }
  session->m_log = log;
  session->m_session->SetLog(log);

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_set_log() - Returning success!");
  return XMA_SUCCESS;
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_send_data(XmaDecoderSession* session, XmaDataBuffer* data, int32_t* data_used) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data()");

  int32_t ret = XMA_SUCCESS;
  if (!data) {
    xma::queue_t queue = xma::QueueDataBuffer();
    ret                = session->m_session->Enqueue(queue);
  } else {
    xma::PrivateDataBufferInfo* info = static_cast<xma::PrivateDataBufferInfo*>(data->private_do_not_touch);
    if ((!info) || (!info->IsValid())) {
      goto fail;
    }
    xma::IntDataBufferPtr buf = info->GetSelf();
    if (!buf) {
      goto fail;
    }
    xma::queue_t queue                         = xma::QueueDataBuffer();
    std::get<xma::QueueDataBuffer>(queue).data = buf;
    ret                                        = session->m_session->Enqueue(queue);
    if (ret != XMA_TRY_AGAIN) {
      if (buf->GetRefCount() == 1) {
        info->SetSelf(nullptr);
      } else {
        buf->DecRef();
      }
    }

    if (data_used) {
      *data_used = data->alloc_size;
    }
  }

  char error_msg[256];
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
  return ret;
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
fail:
  session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_send_data() - Invalid data buffer!");
  return XMA_ERROR_INVALID;
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_get_properties(XmaDecoderSession* session, XmaFrameProperties* fprops) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session || !fprops) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties()");
  xma::FramePropertiesPtr props = session->m_session->GetProps();
  if (props) {
    fprops->format    = props->GetFormat();
    fprops->sw_format = props->GetSwFormat();
    fprops->width     = props->GetWidth();
    fprops->height    = props->GetHeight();
    int32_t planes    = props->GetNumPlanes();
    for (int32_t plane = 0; plane < planes; ++plane) {
      fprops->linesize[plane] = props->GetPlaneStride(plane);
    }

    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties() - Returning success");
    return XMA_SUCCESS;
  } else {
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_get_properties() - Returning send more data");
    return XMA_SEND_MORE_DATA;
  }
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}

///////////////////////////////////////////////////////////////////////////////
int32_t xma_dec_session_recv_frame(XmaDecoderSession* session, XmaFrame* xma_frame) {
  if (!session) {
    return XMA_ERROR_INVALID;
  }

  if (!session->m_session) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  if (!xma_frame || (xma_frame->data[0].buffer_type != NO_BUFFER) || (xma_frame->data[0].refcount != 1)) {
    session->m_log->LogMsg(XMA_ERROR_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() Invalid parameter!");
    return XMA_ERROR_INVALID;
  }

  TRY()
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame()");
  xma::queue_t queue;
  int32_t      ret = session->m_session->Dequeue(queue);
  if (ret != XMA_SUCCESS) {
    char error_msg[256];
    session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() - Returning %s", xma_perror(ret, error_msg, sizeof(error_msg)));
    return ret;
  }

  CopyFrame(session->m_log, xma_frame, std::move(std::get<xma::QueueFrame>(queue).frame));

  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_MOD, "xma_dec_session_recv_frame() - Returning success!");
  return ret;
  CATCH_ERROR(session->m_log, XMA_DECODER_MOD)
}
