// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "xlnx_ml_xma_props.h"

#define ML_MAX_PARAMS          5

/**
 * xlnx_ml_cleanup_props: Free resources allocated in the ml
 * props.
 *
 * @param ml_xma_props: The xma ml properties
 */
void xlnx_ml_cleanup_props(XmaMioFilterProperties* ml_xma_props)
{
    if (ml_xma_props->params) {
        free(ml_xma_props->params);
    }
}


/**
 * xlnx_ml_xma_params_update: Update ml custom params
 *
 * @param ml_props: ML properties
 * @param custom_xma_params: XMA custom parameters
 * @return None
 */
static void xlnx_ml_xma_params_update(XlnxMLProps*       ml_props,
                                        XmaParameter*          custom_xma_params,
                                        uint32_t*              param_cnt)
{
    if (ml_props->model_name) {
        custom_xma_params[*param_cnt].name     = XMA_MIOFILTER_ML_MODEL_NAME;
        custom_xma_params[*param_cnt].type     = XMA_STRING;
        custom_xma_params[*param_cnt].length   = strlen(ml_props->model_name);
        custom_xma_params[*param_cnt].value    = (ml_props->model_name);
        (*param_cnt)++;
    }
    if (ml_props->model_arg) {
        custom_xma_params[*param_cnt].name     = XMA_MIOFILTER_ML_MODEL_ARGS;
        custom_xma_params[*param_cnt].type     = XMA_STRING;
        custom_xma_params[*param_cnt].length   = strlen(ml_props->model_arg);
        custom_xma_params[*param_cnt].value    = (ml_props->model_arg);
        (*param_cnt)++;
    }

    custom_xma_params[*param_cnt].name     = XMA_MIOFILTER_ML_PERFORMANCE;
    custom_xma_params[*param_cnt].type     = XMA_STRING;
    custom_xma_params[*param_cnt].length   = strlen(ml_props->ml_perf);
    custom_xma_params[*param_cnt].value    = (ml_props->ml_perf);
    (*param_cnt)++;

    custom_xma_params[*param_cnt].name     = XMA_MIOFILTER_ML_TENSOR_DUMP;
    custom_xma_params[*param_cnt].type     = XMA_STRING;
    custom_xma_params[*param_cnt].length   = strlen(ml_props->ml_tensor_dump);
    custom_xma_params[*param_cnt].value    = (ml_props->ml_tensor_dump);
    (*param_cnt)++;

    custom_xma_params[*param_cnt].name     = XMA_MIOFILTER_ML_INF_REPEAT;
    custom_xma_params[*param_cnt].type     = XMA_STRING;
    custom_xma_params[*param_cnt].length   = strlen(ml_props->ml_inf_repeat);
    custom_xma_params[*param_cnt].value    = (ml_props->ml_inf_repeat);
    (*param_cnt)++;

    return;
}

/**
 * xlnx_ml_create_props: Fill the input/output fields of the xma
 * ml props.
 *
 * @param param_ctx: The parameter context containing ml param
 *                 values
 * @param ml_xma_props: The ml xma properties which will be
 *                    filled
 * @return XMA_APP_SUCCESS on success
 */
static int32_t xlnx_ml_create_props(XmaHandle            handle,
                                      XlnxMLProps*     param_ctx,
                                      XmaFilterProperties* xma_upload_props,
                                      XmaFilterProperties* xma_download_props,
                                      XmaMioFilterProperties* ml_xma_props)
{
    // setup frame poperties
    ml_xma_props->hwfilter_type        = XMA_ML_FILTER_TYPE;
    ml_xma_props->param_cnt            = 0;
    ml_xma_props->params               = (XmaParameter*)calloc(1, ML_MAX_PARAMS * sizeof(XmaParameter));
    ml_xma_props->num_inputs          = param_ctx->num_inputs;
    ml_xma_props->inputs[0].format         = XMA_VPE_FMT_TYPE;
    ml_xma_props->inputs[0].sw_format      = param_ctx->input_pix_fmt;
    ml_xma_props->inputs[0].width          = param_ctx->input_width;
    ml_xma_props->inputs[0].height         = param_ctx->input_height;
    ml_xma_props->inputs[0].stride         = param_ctx->input_width;
    ml_xma_props->inputs[0].framerate.numerator   = param_ctx->input_fps_num;
    ml_xma_props->inputs[0].framerate.denominator = param_ctx->input_fps_den;
 
    ml_xma_props->num_outputs          = param_ctx->num_outputs;
    ml_xma_props->outputs[0].format         = XMA_VPE_FMT_TYPE;
    ml_xma_props->outputs[0].sw_format      = param_ctx->input_pix_fmt;
    ml_xma_props->outputs[0].width          = param_ctx->input_width;
    ml_xma_props->outputs[0].height         = param_ctx->input_height;
    ml_xma_props->outputs[0].stride         = param_ctx->input_width;
    ml_xma_props->outputs[0].framerate.numerator   = param_ctx->input_fps_num;
    ml_xma_props->outputs[0].framerate.denominator = param_ctx->input_fps_den;

    xma_upload_props->hwfilter_type = XMA_UPLOAD_FILTER_TYPE;
    xma_upload_props->handle        = handle;
    xma_upload_props->param_cnt     = 0;
    xma_upload_props->params        = NULL;
    xma_upload_props->input.sw_format = ml_xma_props->inputs[0].sw_format;
    xma_upload_props->input.format    = xma_upload_props->input.sw_format;
    xma_upload_props->input.width     = ml_xma_props->inputs[0].width;
    xma_upload_props->input.height    = ml_xma_props->inputs[0].height;
    xma_upload_props->input.framerate.numerator =
        ml_xma_props->inputs[0].framerate.numerator;
    xma_upload_props->input.framerate.denominator =
        ml_xma_props->inputs[0].framerate.denominator;
    xma_upload_props->output.format = XMA_VPE_FMT_TYPE;
    xma_upload_props->output.sw_format = ml_xma_props->inputs[0].sw_format;
    xma_upload_props->output.width     = ml_xma_props->inputs[0].width;
    xma_upload_props->output.height    = ml_xma_props->inputs[0].height;
    xma_upload_props->output.framerate.numerator =
        ml_xma_props->inputs[0].framerate.numerator;
    xma_upload_props->output.framerate.denominator =
        ml_xma_props->inputs[0].framerate.denominator;

    xma_download_props->hwfilter_type        = XMA_DOWNLOAD_FILTER_TYPE;
    xma_download_props->handle               = handle;
    xma_download_props->param_cnt            = 0;
    xma_download_props->params               = NULL;
    xma_download_props->input.format         = XMA_VPE_FMT_TYPE;
    xma_download_props->input.sw_format      = ml_xma_props->inputs[0].sw_format;
    xma_download_props->input.width          = ml_xma_props->inputs[0].width;
    xma_download_props->input.height         = ml_xma_props->inputs[0].height;
    xma_download_props->input.framerate.numerator =
        ml_xma_props->inputs[0].framerate.numerator;
    xma_download_props->input.framerate.denominator =
        ml_xma_props->inputs[0].framerate.denominator;
    xma_download_props->output.sw_format      = ml_xma_props->inputs[0].sw_format;
    xma_download_props->output.format         = xma_download_props->output.sw_format;
    xma_download_props->output.width          = ml_xma_props->inputs[0].width;
    xma_download_props->output.height         = ml_xma_props->inputs[0].height;
    xma_download_props->output.framerate.numerator =
        ml_xma_props->inputs[0].framerate.numerator;
    xma_download_props->output.framerate.denominator =
        ml_xma_props->inputs[0].framerate.denominator;

    xlnx_ml_xma_params_update(param_ctx, ml_xma_props->params,
                               &ml_xma_props->param_cnt);

    return XMA_APP_SUCCESS;
}

/**
 * xlnx_ml_create_xma_props: Create the xma ml properties
 * given the param ctx and session id.
 *
 * @param handle: The xma handle shared between xma sessions.
 * @param param_ctx: The parameter context containing ml param
 *                 values
 * @param ml_xma_props: The ml xma properties which will be
 *                        filled
 * @return XMA_APP_SUCCESS on success
 */
int32_t xlnx_ml_create_xma_props(XmaHandle handle, XlnxMLProps* param_ctx,
                                   XmaFilterProperties* xma_upload_props,
                                   XmaFilterProperties* xma_download_props,
                                   XmaMioFilterProperties* ml_xma_props)
{
    // Setup custom xma plugin params
    xlnx_ml_create_props(handle, param_ctx, xma_upload_props,
                           xma_download_props, ml_xma_props);
    return XMA_APP_SUCCESS;
}
